import SocketServer
import os
import shutil
import socket
import struct
import tempfile
import threading
import unittest

from dyntftpd.server import TFTPServer, FileSystemHandler, TFTPUDPHandler


class ThreadedTFTPServer(SocketServer.ThreadingMixIn, TFTPServer):
    pass


class TFTPServerTestCase(unittest.TestCase):

    def setUp(self, handler=FileSystemHandler):
        """ Starts an instance of TFTPServer and initializes a client socket.
        """
        self.tftp_root = tempfile.mkdtemp()

        self.server = ThreadedTFTPServer(
            host='127.0.0.1', port=0, root=self.tftp_root, handler=handler
        )
        self.listen_ip, self.listen_port = self.server.socket.getsockname()

        self.server_thread = threading.Thread(target=self.server.serve_forever)
        self.server_thread.start()

        self.client_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def send(self, message):
        self.client_socket.sendto(message, (self.listen_ip, self.listen_port))

    def get_file(self, filename, mode='octet', options=None):
        request = '\x00\x01%s\x00%s\x00' % (filename, mode)
        if options is not None:
            request += ''.join([
                '%s\x00%s\x00' % (k, v) for k, v in options.items()
            ])
        return self.send(request)

    def recv(self, size=1024):
        data, addr = self.client_socket.recvfrom(size)
        return data, addr

    def ack(self, block_string):
        return self.send('\x00\x04' + block_string)

    def ack_n(self, block_id):
        return self.ack(struct.pack('>h', block_id))

    def tearDown(self):
        shutil.rmtree(self.tftp_root)
        self.server.shutdown()
        self.server.server_close()


class TestFileSystemHandler(TFTPServerTestCase):

    def test_invalid_request(self):
        # Missing trailing \x00
        self.send('\x00\x01filename\x00octet')
        data, _ = self.recv()
        self.assertTrue(data.startswith('\x00\x05\x00\x04'))

        # Not enough arguments
        self.send('\x00\x01')
        data, _ = self.recv()
        self.assertTrue(data.startswith('\x00\x05\x00\x04'))

        self.send('\x00\x01filename\x00')
        data, _ = self.recv()
        self.assertTrue(data.startswith('\x00\x05\x00\x04'))

    def test_non_existing(self):
        self.get_file('invalid')
        data, _ = self.recv()
        # \x00\x05 = error
        # \x00\x01 = no such file
        self.assertTrue(data.startswith('\x00\x05\x00\x01'))

    def test_non_octet(self):
        """ Currently, the server only supports ascii transfers.
        """
        self.get_file('yo.txt', mode='ascii')
        data, _ = self.recv()
        # \x00\x05 = error
        # \x00\x04 = illegal operation
        self.assertTrue(data.startswith('\x00\x05\x00\x04'))

    def test_empty_file(self):
        handle = open(os.path.join(self.tftp_root, 'test.txt'), 'w+')
        handle.close()

        self.get_file('test.txt')
        data, _ = self.recv()
        # \x00\x03 = data
        # \x00\x01 = block id 1
        self.assertEqual(data, '\x00\x03\x00\x01')
        self.ack_n(1)

    def test_small_file(self):
        handle = open(os.path.join(self.tftp_root, 'test.txt'), 'w+')
        handle.write('hello world')
        handle.flush()

        self.get_file('test.txt')
        data, _ = self.recv()
        # \x00\x03 = data
        # \x00\x01 = block id 1
        self.assertEqual(data, '\x00\x03\x00\x01hello world')
        self.ack_n(1)

    def test_options(self):
        """ Ensure blksize can be given, and other options are ignored.

        http://tools.ietf.org/html/rfc1782
        """
        handle = open(os.path.join(self.tftp_root, 'test.txt'), 'w+')
        handle.write('hello world')
        handle.flush()

        self.get_file('test.txt', options={'blksize': 1024, 'timeout': 5})
        data, _ = self.recv()
        # \x00\x06 = OACK
        self.assertEqual(data, '\x00\x06blksize\x001024\x00')

    def test_big_file(self):
        handle = open(os.path.join(self.tftp_root, 'test.txt'), 'w+')
        handle.write('A' * 512)
        handle.write('B' * 512)
        handle.write('C' * 32)
        handle.flush()

        self.get_file('test.txt')
        data, _ = self.recv()
        # \x00\x03 = data
        # \x00\x01 = block id 1
        self.assertEqual(len(data), 516)
        self.assertEqual(data, '\x00\x03\x00\x01' + 'A' * 512)
        self.ack_n(1)

        data, _ = self.recv()
        self.assertEqual(len(data), 516)
        self.assertEqual(data, '\x00\x03\x00\x02' + 'B' * 512)
        self.ack_n(2)

        data, _ = self.recv()
        self.assertEqual(len(data), 36)
        self.assertEqual(data, '\x00\x03\x00\x03' + 'C' * 32)
        self.ack_n(3)

    def test_big_file_with_blksize(self):
        handle = open(os.path.join(self.tftp_root, 'test.txt'), 'w+')
        handle.write('A' * 512)
        handle.write('B' * 512)
        handle.write('C' * 32)
        handle.flush()

        self.get_file('test.txt', options={'blksize': 1024})

        # \x00\x06 = OACK
        data, _ = self.recv()
        self.assertEqual(data, '\x00\x06blksize\x001024\x00')
        self.ack_n(0)

        data, _ = self.recv(size=4096)
        self.assertEqual(len(data), 1024 + 4)
        # \x00\x03 = data
        # \x00\x01 = block id 1
        self.assertEqual(data, '\x00\x03\x00\x01' + 'A' * 512 + 'B' * 512)
        self.ack_n(1)

        data, _ = self.recv(size=4096)
        self.assertEqual(len(data), 32 + 4)
        # \x00\x03 = data
        # \x00\x01 = block id 2
        self.assertEqual(data, '\x00\x03\x00\x02' + 'C' * 32)

    def test_directory_transversal(self):
        self.get_file('../../yo.txt')
        data, _ = self.recv()
        # \x00\x05 = error
        # \x00\x04 = permission denied
        self.assertTrue(data.startswith('\x00\x05\x00\x02'))

    def test_permission_denied(self):
        filename = os.path.join(self.tftp_root, 'test.txt')
        handle = open(filename, 'w+')
        handle.write('hello world')
        os.chmod(filename, 0000)
        handle.flush()

        self.get_file('test.txt')
        data, _ = self.recv()
        # \x00\x05 = error
        # \x00\x04 = permission denied
        self.assertTrue(data.startswith('\x00\x05\x00\x02'))

    def test_retransmission(self):
        handle = open(os.path.join(self.tftp_root, 'test.txt'), 'w+')
        handle.write('A' * 512)
        handle.write('B' * 512)
        handle.flush()

        self.get_file('test.txt')
        data, _ = self.recv()
        # \x00\x03 = data
        # \x00\x01 = block id 1
        self.assertEqual(len(data), 516)
        self.assertEqual(data, '\x00\x03\x00\x01' + 'A' * 512)
        self.ack_n(1)

        data, _ = self.recv()
        self.assertEqual(len(data), 516)
        self.assertEqual(data, '\x00\x03\x00\x02' + 'B' * 512)

        # Retransmit. Aknowledge the last packet.
        self.ack_n(1)
        data, _ = self.recv()
        self.assertEqual(len(data), 516)
        self.assertEqual(data, '\x00\x03\x00\x02' + 'B' * 512)


class CustomHandler(TFTPUDPHandler):

    def sanitize_filename(self, filename):
        return filename

    def load_file(self, filename):
        raise OSError('xxx')


class TestCustomHandler(TFTPServerTestCase):

    def setUp(self):
        return super(TestCustomHandler, self).setUp(handler=CustomHandler)

    def test_load_file_failing(self):
        self.get_file('whatever')
        data, _ = self.recv()
        # \x00\x05 = error
        # \x00\x00 = undefined
        self.assertEqual(data, '\x00\x05\x00\x00Internal error\x00')



import logging


__version__ = '0.2.0'

# Prevent message "No handlers could be found for logger "dyntftpd"" to be
# displayed if not logger is setup by the application.
logging.getLogger(__name__).addHandler(logging.NullHandler())


from .server import TFTPServer, FileSystemHandler



import argparse
import logging

from .server import TFTPServer


def arguments_parser():
    parser = argparse.ArgumentParser(
        description='Extendable TFTP server, implemented in Python'
    )
    parser.add_argument('--host', '-H', default='')
    parser.add_argument('--port', '-p', default=69, type=int)
    parser.add_argument(
        '--root', '-r', default='/var/lib/tftpboot/', help='TFTP root folder'
    )
    return parser


def main():
    parser = arguments_parser()
    parser.add_argument(
        '-v', dest='verbose', action='count', default=0, help='Verbose mode'
    )
    args = parser.parse_args()

    log_level = logging.DEBUG if args.verbose else logging.INFO
    logging.basicConfig(
        level=log_level,
        format='[%(asctime)-15s] %(client_ip)s: %(message)s'
    )

    tftp_server = TFTPServer(args.host, args.port, root=args.root)
    tftp_server.serve_forever()



import os
import re

from setuptools import setup, find_packages


MODULE_NAME = 'dyntftpd'

DEPENDENCIES = []
TEST_DEPENDENCIES = []


def find_file(*paths):
    return os.path.join(os.path.dirname(__file__), *paths)


def get_version():
    """ Reads package version number from package's __init__.py. """
    with open(find_file(MODULE_NAME, '__init__.py')) as init:
        for line in init.readlines():
            res = re.match(r'^__version__ = [\'"](.*)[\'"]$', line)
            if res:
                return res.group(1)

        raise NotImplementedError(
            '%s does not have a __version__ defined in '
            '__init__.py' % MODULE_NAME
        )


def get_long_description():
    """ Reads description from README and CHANGES. """
    with open(find_file('README.rst')) as readme, \
            open(find_file('CHANGES.rst')) as changes:

        return readme.read() + '\n' + changes.read()


VERSION = get_version()


setup(
    name=MODULE_NAME,
    version=VERSION,
    description='A simple TFTP server',
    download_url='https://github.com/brmzkw/dyntftpd/tarball/v' + VERSION,
    long_description=get_long_description(),
    author='Julien Castets',
    author_email='castets.j@gmail.com',
    url='https://github.com/brmzkw/dyntftpd',
    packages=find_packages(),
    install_requires=DEPENDENCIES,
    tests_require=DEPENDENCIES + TEST_DEPENDENCIES,
    test_suite='tests',
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2 :: Only',  # fuck python3
        'Topic :: Internet :: File Transfer Protocol (FTP)'
    ],
    entry_points={
        'console_scripts': [
            'dyntftpd = dyntftpd.cli:main'
        ]
    }
)



import errno
import logging
import os
import struct

import SocketServer


logger = logging.getLogger(__name__)


class TFTPUDPHandler(SocketServer.BaseRequestHandler):
    """ Mixin. Implementation of the TFTP protocol.

    http://tools.ietf.org/html/rfc1350
    http://tools.ietf.org/html/rfc1782
    """

    # opcodes
    OP_RRQ = 1
    OP_WRQ = 2
    OP_DATA = 3
    OP_ACK = 4
    OP_ERROR = 5
    OP_OACK = 6

    # errors
    ERR_UNDEFINED = 0
    ERR_NOT_FOUND = 1
    ERR_PERM = 2
    ERR_DISK_FULL = 3
    ERR_ILLEGAL_OPERATION = 4
    ERR_UNKNOWN_TID = 5
    ERR_ALREADY_EXISTS = 6
    ERR_NO_SUCH_USER = 7

    def _log(self, level, msg, extra=None, exc_info=False):
        """ Add client_ip to extra.
        """
        log_extra = {'client_ip': self.client_address[0]}
        if extra:
            log_extra.update(extra)
        logger.log(level, msg, extra=log_extra, exc_info=exc_info)

    def handle(self):
        """ Extract header info and dispatch to handle_* methods.
        """
        data = self.request[0]
        opcode, = struct.unpack('!h', data[0:2])

        data = data[2:]  # skip opcode

        if opcode == self.OP_RRQ:
            args = data.split('\x00')

            if args[-1] != '':
                return self.send_error(
                    self.ERR_ILLEGAL_OPERATION,
                    'Final argument should end with a \\0'
                )
            args = args[:-1]  # skip final \0

            if len(args) < 2:
                return self.send_error(
                    self.ERR_ILLEGAL_OPERATION,
                    "Filename and mode required"
                )

            filename, mode = args[0], args[1]
            options = args[2:]

            # options must be a list like: [optname1, optvalue1, ..., optnameN,
            # optvalueN)
            if len(options) % 2:
                return self.send_error(
                    self.ERR_ILLEGAL_OPERATION,
                    "Malformed options"
                )

            # transform options to a dict
            options = dict(options[i:i+2] for i in xrange(0, len(options), 2))

            self.handle_rrq(filename, mode, options)

        elif opcode == self.OP_ACK:
            block_id, = struct.unpack('!h', data)
            self.handle_ack(block_id)

        else:
            return self.send_error(
                self.ERR_ILLEGAL_OPERATION,
                'Opcode %d not handled by the server' % opcode
            )

    def handle_rrq(self, filename, mode, options):
        """ Handle READ requests.

        Create a new session.
        """
        self._log(logging.INFO, 'GET %s (%s)' % (filename, mode))

        if mode.lower() != 'octet':
            self.send_error(
                self.ERR_ILLEGAL_OPERATION,
                'Only octet mode is supported by the server'
            )
            return

        try:
            filename = self.sanitize_filename(filename)
        except ValueError as exc:
            self.send_error(self.ERR_PERM, str(exc))
            return

        try:
            handle = self.load_file(filename)
        except IOError as exc:
            # If ENOENT, consider the file is missing. Otherwise, consider we
            # don't have the permission to read it.
            err_msg = exc.strerror or str(exc)
            if exc.errno == errno.ENOENT:
                self.send_error(
                    self.ERR_NOT_FOUND, '%s (%s)' % (err_msg, filename)
                )
            else:
                self.send_error(
                    self.ERR_PERM, '%s (%s)' % (err_msg, filename)
                )
            return
        # The file cannot be loaded for any (critical) reason. Log the
        # traceback.
        except Exception as exc:
            self._log(logging.ERROR, 'Internal error', exc_info=True)
            self.send_error(self.ERR_UNDEFINED, 'Internal error')
            return

        session =  TFTPSession(filename, handle)
        self.server.sessions[self.client_address] = session

        # If there is a supported option, return a OACK, otherwise return the
        # first packet.
        # For now, only 'blksize' is supported.
        blksize = options.get('blksize')

        # Set the block size and return a OACK
        if blksize:
            try:
                session.blksize = int(blksize)
            except ValueError:  # not an int
                return self.send_error(
                    self.ERR_ILLEGAL_OPERATION, 'Bad option value'
                )
            return self.send_oack(blksize=blksize)

        # No options, return the first part of the file
        self.send_data()

    def sanitize_filename(self, filename):
        """ Compute a filename that can be loaded by load_file. Make security
        checks to prevent path transversal.
        """
        raise NotImplementedError()

    def load_file(self, filename):
        """ Return a FILE like object.
        """
        raise NotImplementedError()

    def handle_ack(self, block_id):
        """ Client has aknowledged a block id. Can be a retransmission or the
        next packet to send.
        """
        self._log(logging.DEBUG, 'ACK (block %s)' % block_id)
        session = self.server.sessions.get(self.client_address)

        # If the ACK does not correspond to a read request
        if not session:
            return

        # Last packet was received
        if block_id == session.block_id + 1:

            # Final ACK from the client, kill the sesssion
            if session.last_read_is_eof:
                self._log(
                    logging.INFO,
                    'Transfer of %s successful' % session.filename
                )
                del self.server.sessions[self.client_address]
                return

            # Next packet
            session.block_id += 1

        # Send the next packet, or retransmit the last packet if there was an
        # error
        self.send_data()

    def send_oack(self, **options):
        """ Send options acknowledgement.
        """
        packed = struct.pack('!h', self.OP_OACK)
        for key, value in options.iteritems():
            packed += key + '\x00' + value + '\x00'

        socket = self.request[1]
        socket.sendto(packed, self.client_address)

    def send_data(self):
        """ Send the next data packet to the client.
        """
        session = self.server.sessions[self.client_address]
        session.handle.seek(session.block_id * session.blksize)
        data = session.handle.read(session.blksize)
        session.last_read_is_eof = len(data) < session.blksize

        packed = struct.pack('!hh', self.OP_DATA, session.block_id + 1)
        packed += data

        socket = self.request[1]
        socket.sendto(packed, self.client_address)

    def send_error(self, error_code, error_msg):
        """ Send error packet to the client.
        """
        self._log(logging.ERROR, error_msg)
        socket = self.request[1]
        packed = struct.pack('!hh', self.OP_ERROR, error_code)
        packed += error_msg + '\x00'
        socket.sendto(packed, self.client_address)


class TFTPSession(object):

    def __init__(self, filename, handle):
        self.block_id = 0
        self.filename = filename
        self.handle = handle
        self.last_read_is_eof = False
        self.blksize = 512


class FileSystemHandler(TFTPUDPHandler):

    def sanitize_filename(self, filename):
        """ Raise if trying to open a file up to the root folder.
        """
        server_root = os.path.abspath(self.server.root)
        abs_path = os.path.abspath(os.path.join(server_root, filename))
        if os.path.commonprefix([abs_path, server_root]) != server_root:
            raise ValueError('Directory trasversal prevented')
        return abs_path

    def load_file(self, filename):
        return open(filename)


class TFTPServer(SocketServer.ThreadingMixIn, SocketServer.UDPServer):

    def __init__(self, host='', port=69, root='/var/lib/tftpboot',
                 handler=FileSystemHandler):

        self.sessions = {}
        self.root = root
        SocketServer.UDPServer.__init__(self, (host, port), handler)
